import json
import re
from json.decoder import JSONDecodeError, JSONDecoder, scanstring

from apistar.exceptions import ErrorMessage, ParseError, Position
from apistar.tokenize.tokens import DictToken, ListToken, ScalarToken

FLAGS = re.VERBOSE | re.MULTILINE | re.DOTALL
WHITESPACE = re.compile(r'[ \t\n\r]*', FLAGS)
WHITESPACE_STR = ' \t\n\r'
NUMBER_RE = re.compile(
    r'(-?(?:0|[1-9]\d*))(\.\d+)?([eE][-+]?\d+)?',
    (re.VERBOSE | re.MULTILINE | re.DOTALL))


def _TokenizingJSONObject(s_and_end, strict, scan_once,
                          memo, content, _w=WHITESPACE.match, _ws=WHITESPACE_STR):
    s, end = s_and_end
    pairs = []
    pairs_append = pairs.append
    memo_get = memo.setdefault
    # Use a slice to prevent IndexError from being raised, the following
    # check will raise a more specific ValueError if the string is empty
    nextchar = s[end:end + 1]
    # Normally we expect nextchar == '"'
    if nextchar != '"':
        if nextchar in _ws:
            end = _w(s, end).end()
            nextchar = s[end:end + 1]
        # Trivial empty object
        if nextchar == '}':
            return {}, end + 1
        elif nextchar != '"':
            raise JSONDecodeError(
                "Expecting property name enclosed in double quotes", s, end)
    end += 1
    while True:
        start = end - 1
        key, end = scanstring(s, end, strict)
        key = memo_get(key, key)
        key = ScalarToken(memo_get(key, key), start, end - 1, content)
        # To skip some function call overhead we optimize the fast paths where
        # the JSON key separator is ": " or just ":".
        if s[end:end + 1] != ':':
            end = _w(s, end).end()
            if s[end:end + 1] != ':':
                raise JSONDecodeError("Expecting ':' delimiter", s, end)
        end += 1

        try:
            if s[end] in _ws:
                end += 1
                if s[end] in _ws:
                    end = _w(s, end + 1).end()
        except IndexError:
            pass

        try:
            value, end = scan_once(s, end)
        except StopIteration as err:
            raise JSONDecodeError("Expecting value", s, err.value) from None
        pairs_append((key, value))
        try:
            nextchar = s[end]
            if nextchar in _ws:
                end = _w(s, end + 1).end()
                nextchar = s[end]
        except IndexError:
            nextchar = ''
        end += 1

        if nextchar == '}':
            break
        elif nextchar != ',':
            raise JSONDecodeError("Expecting ',' delimiter", s, end - 1)
        end = _w(s, end).end()
        nextchar = s[end:end + 1]
        end += 1
        if nextchar != '"':
            raise JSONDecodeError(
                "Expecting property name enclosed in double quotes", s, end - 1)
    return dict(pairs), end


def _make_scanner(context, content):
    parse_object = _TokenizingJSONObject
    parse_array = context.parse_array
    parse_string = context.parse_string
    match_number = NUMBER_RE.match
    strict = context.strict
    parse_float = context.parse_float
    parse_int = context.parse_int
    memo = context.memo

    def _scan_once(string, idx):
        try:
            nextchar = string[idx]
        except IndexError:
            raise StopIteration(idx) from None

        if nextchar == '"':
            value, end = parse_string(string, idx + 1, strict)
            return ScalarToken(value, idx, end - 1, content), end
        elif nextchar == '{':
            value, end = parse_object(
                (string, idx + 1), strict,
                _scan_once, memo, content
            )
            return DictToken(value, idx, end - 1, content), end
        elif nextchar == '[':
            value, end = parse_array((string, idx + 1), _scan_once)
            return ListToken(value, idx, end - 1, content), end
        elif nextchar == 'n' and string[idx:idx + 4] == 'null':
            value, end = None, idx + 4
            return ScalarToken(value, idx, end - 1, content), end
        elif nextchar == 't' and string[idx:idx + 4] == 'true':
            value, end = True, idx + 4
            return ScalarToken(value, idx, end - 1, content), end
        elif nextchar == 'f' and string[idx:idx + 5] == 'false':
            value, end = False, idx + 5
            return ScalarToken(value, idx, end - 1, content), end

        m = match_number(string, idx)
        if m is not None:
            integer, frac, exp = m.groups()
            if frac or exp:
                res = parse_float(integer + (frac or '') + (exp or ''))
            else:
                res = parse_int(integer)
            value, end = res, m.end()
            return ScalarToken(value, idx, end - 1, content), end
        else:
            raise StopIteration(idx)

    def scan_once(string, idx):
        try:
            return _scan_once(string, idx)
        finally:
            memo.clear()

    return scan_once


class _TokenizingDecoder(JSONDecoder):
    def __init__(self, *args, **kwargs):
        content = kwargs.pop('content')
        super().__init__(*args, **kwargs)
        self.scan_once = _make_scanner(self, content)


def _strip_endings(text, endings):
    for ending in endings:
        if text.endswith(ending):
            return text[:-len(ending)]
    return text


def tokenize_json(content):
    assert isinstance(content, (str, bytes))

    if isinstance(content, bytes):
        content = content.decode('utf-8', 'ignore')

    if not content.strip():
        message = ErrorMessage(
            text='No content.',
            code='parse_error',
            position=Position(line_no=1, column_no=1, index=0)
        )
        raise ParseError(messages=[message], summary='Invalid JSON.')

    try:
        decoder = _TokenizingDecoder(content=content)
        return decoder.decode(content)
    except json.decoder.JSONDecodeError as exc:
        message = ErrorMessage(
            text=_strip_endings(exc.msg, [" starting at", " at"]) + ".",
            code='parse_error',
            position=Position(line_no=exc.lineno, column_no=exc.colno, index=exc.pos)
        )
        raise ParseError(messages=[message], summary='Invalid JSON.') from None



import re
import typing
from math import isfinite

from apistar import formats
from apistar.compat import dict_type
from apistar.exceptions import ErrorMessage, ValidationError

NO_DEFAULT = object()

FORMATS = {
    'date': formats.DateFormat(),
    'time': formats.TimeFormat(),
    'datetime': formats.DateTimeFormat()
}


class Validator:
    errors = {}
    _creation_counter = 0

    def __init__(self, title='', description='', default=NO_DEFAULT, allow_null=False, definitions=None, def_name=None):
        definitions = {} if (definitions is None) else dict_type(definitions)

        assert isinstance(title, str)
        assert isinstance(description, str)
        assert isinstance(definitions, dict)
        assert all(isinstance(k, str) for k in definitions.keys())
        assert all(isinstance(v, Validator) for v in definitions.values())

        if allow_null and default is NO_DEFAULT:
            default = None

        if default is not NO_DEFAULT:
            self.default = default

        self.title = title
        self.description = description
        self.allow_null = allow_null
        self.definitions = definitions
        self.def_name = def_name

        # We need this global counter to determine what order fields have
        # been declared in when used with `Type`.
        self._creation_counter = Validator._creation_counter
        Validator._creation_counter += 1

    def validate(self, value, definitions=None, allow_coerce=False):
        raise NotImplementedError()

    def is_valid(self, value):
        try:
            self.validate(value)
        except ValidationError:
            return False
        return True

    def has_default(self):
        return hasattr(self, 'default')

    def error(self, code, **context):
        message = self.error_message(code, **context)
        raise ValidationError(messages=[message])

    def error_message(self, code, **context):
        text = self.errors[code].format(**self.__dict__, **context)
        return ErrorMessage(text=text, code=code)

    def get_definitions(self, definitions=None):
        if self.definitions is None and self.def_name is None:
            return definitions

        if definitions is None:
            definitions = {}
        if self.definitions is not None:
            definitions.update(self.definitions)
        if self.def_name is not None:
            definitions[self.def_name] = self
        return definitions

    def __or__(self, other):
        if isinstance(self, Union):
            items = self.items
        else:
            items = [self]

        if isinstance(other, Union):
            items += other.items
        else:
            items += [other]

        return Union(items)


class String(Validator):
    errors = {
        'type': 'Must be a string.',
        'null': 'May not be null.',
        'blank': 'Must not be blank.',
        'max_length': 'Must have no more than {max_length} characters.',
        'min_length': 'Must have at least {min_length} characters.',
        'pattern': 'Must match the pattern /{pattern}/.',
        'format': 'Must be a valid {format}.',
        'enum': 'Must be one of {enum}.',
        'exact': 'Must be {exact}.'
    }

    def __init__(self, max_length=None, min_length=None, pattern=None,
                 enum=None, format=None, **kwargs):
        super().__init__(**kwargs)

        assert max_length is None or isinstance(max_length, int)
        assert min_length is None or isinstance(min_length, int)
        assert pattern is None or isinstance(pattern, str)
        assert enum is None or isinstance(enum, list) and all([isinstance(i, str) for i in enum])
        assert format is None or isinstance(format, str)

        self.max_length = max_length
        self.min_length = min_length
        self.pattern = pattern
        self.enum = enum
        self.format = format

    def validate(self, value, definitions=None, allow_coerce=False):
        if value is None and self.allow_null:
            return None
        elif value is None:
            self.error('null')
        elif self.format in FORMATS and FORMATS[self.format].is_native_type(value):
            return value
        elif not isinstance(value, str):
            self.error('type')

        if self.enum is not None:
            if value not in self.enum:
                if len(self.enum) == 1:
                    self.error('exact', exact=self.enum[0])
                self.error('enum')

        if self.min_length is not None:
            if len(value) < self.min_length:
                if self.min_length == 1:
                    self.error('blank')
                else:
                    self.error('min_length')

        if self.max_length is not None:
            if len(value) > self.max_length:
                self.error('max_length')

        if self.pattern is not None:
            if not re.search(self.pattern, value):
                self.error('pattern')

        if self.format in FORMATS:
            return FORMATS[self.format].validate(value)

        return value


class NumericType(Validator):
    """
    Base class for both `Number` and `Integer`.
    """
    numeric_type = None  # type: type
    errors = {
        'type': 'Must be a number.',
        'null': 'May not be null.',
        'integer': 'Must be an integer.',
        'finite': 'Must be finite.',
        'minimum': 'Must be greater than or equal to {minimum}.',
        'exclusive_minimum': 'Must be greater than {minimum}.',
        'maximum': 'Must be less than or equal to {maximum}.',
        'exclusive_maximum': 'Must be less than {maximum}.',
        'multiple_of': 'Must be a multiple of {multiple_of}.',
        'enum': 'Must be one of {enum}.',
        'exact': 'Must be {exact}.'
    }

    def __init__(self, minimum=None, maximum=None, exclusive_minimum=False,
                 exclusive_maximum=False, multiple_of=None, enum=None,
                 format=None, **kwargs):
        super().__init__(**kwargs)

        assert minimum is None or isinstance(minimum, (int, float))
        assert maximum is None or isinstance(maximum, (int, float))
        assert isinstance(exclusive_minimum, bool)
        assert isinstance(exclusive_maximum, bool)
        assert multiple_of is None or isinstance(multiple_of, (int, float))
        assert enum is None or isinstance(enum, list) and all([isinstance(i, str) for i in enum])
        assert format is None or isinstance(format, str)

        self.minimum = minimum
        self.maximum = maximum
        self.exclusive_minimum = exclusive_minimum
        self.exclusive_maximum = exclusive_maximum
        self.multiple_of = multiple_of
        self.enum = enum
        self.format = format

    def validate(self, value, definitions=None, allow_coerce=False):
        if value is None and self.allow_null:
            return None
        elif value is None:
            self.error('null')
        elif isinstance(value, bool):
            self.error('type')
        elif self.numeric_type is int and isinstance(value, float) and not value.is_integer():
            self.error('integer')
        elif not isinstance(value, (int, float)) and not allow_coerce:
            self.error('type')
        elif isinstance(value, float) and not isfinite(value):
            self.error('finite')

        try:
            value = self.numeric_type(value)
        except (TypeError, ValueError):
            self.error('type')

        if self.enum is not None:
            if value not in self.enum:
                if len(self.enum) == 1:
                    self.error('exact', exact=self.enum[0])
                self.error('enum')

        if self.minimum is not None:
            if self.exclusive_minimum:
                if value <= self.minimum:
                    self.error('exclusive_minimum')
            else:
                if value < self.minimum:
                    self.error('minimum')

        if self.maximum is not None:
            if self.exclusive_maximum:
                if value >= self.maximum:
                    self.error('exclusive_maximum')
            else:
                if value > self.maximum:
                    self.error('maximum')

        if self.multiple_of is not None:
            if isinstance(self.multiple_of, float):
                if not (value * (1 / self.multiple_of)).is_integer():
                    self.error('multiple_of')
            else:
                if value % self.multiple_of:
                    self.error('multiple_of')

        return value


class Number(NumericType):
    numeric_type = float


class Integer(NumericType):
    numeric_type = int


class Boolean(Validator):
    errors = {
        'type': 'Must be a valid boolean.',
        'null': 'May not be null.',
    }
    values = {
        'true': True,
        'false': False,
        'on': True,
        'off': False,
        '1': True,
        '0': False,
        '': False,
    }
    null_values = {
        '': None,
        'null': None,
        'none': None,
    }

    def validate(self, value, definitions=None, allow_coerce=False):
        if value is None and self.allow_null:
            return None

        elif value is None:
            self.error('null')

        elif not isinstance(value, bool):
            if allow_coerce and isinstance(value, str):
                if self.allow_null:
                    values = dict(self.value)
                    values.update(self.null_values)
                else:
                    values = self.values
                try:
                    return values[value.lower()]
                except KeyError:
                    pass
            self.error('type')

        return value


class Object(Validator):
    errors = {
        'type': 'Must be an object.',
        'null': 'May not be null.',
        'invalid_key': 'Object keys must be strings.',
        'required': 'The "{field_name}" field is required.',
        'invalid_property': 'Invalid property name.',
        'empty': 'Must not be empty.',
        'max_properties': 'Must have no more than {max_properties} properties.',
        'min_properties': 'Must have at least {min_properties} properties.',
    }

    def __init__(self, properties=None, pattern_properties=None,
                 additional_properties=True, min_properties=None,
                 max_properties=None, required=None,
                 **kwargs):
        super().__init__(**kwargs)

        properties = {} if (properties is None) else dict_type(properties)
        pattern_properties = {} if (pattern_properties is None) else dict_type(pattern_properties)
        required = list(required) if isinstance(required, (list, tuple)) else required
        required = [] if (required is None) else required

        assert all(isinstance(k, str) for k in properties.keys())
        assert all(hasattr(v, 'validate') for v in properties.values())
        assert all(isinstance(k, str) for k in pattern_properties.keys())
        assert all(hasattr(v, 'validate') for v in pattern_properties.values())
        assert additional_properties in (None, True, False) or hasattr(additional_properties, 'validate')
        assert min_properties is None or isinstance(min_properties, int)
        assert max_properties is None or isinstance(max_properties, int)
        assert all(isinstance(i, str) for i in required)

        self.properties = properties
        self.pattern_properties = pattern_properties
        self.additional_properties = additional_properties
        self.min_properties = min_properties
        self.max_properties = max_properties
        self.required = required

    def validate(self, value, definitions=None, allow_coerce=False):
        if value is None and self.allow_null:
            return None
        elif value is None:
            self.error('null')
        elif not isinstance(value, (dict, typing.Mapping)):
            self.error('type')

        definitions = self.get_definitions(definitions)
        validated = dict_type()

        # Ensure all property keys are strings.
        errors = {}
        for key in value.keys():
            if not isinstance(key, str):
                errors[key] = [self.error_message('invalid_key')]

        # Min/Max properties
        if self.min_properties is not None:
            if len(value) < self.min_properties:
                if self.min_properties == 1:
                    self.error('empty')
                else:
                    self.error('min_properties')
        if self.max_properties is not None:
            if len(value) > self.max_properties:
                self.error('max_properties')

        # Required properties
        for key in self.required:
            if key not in value:
                errors[key] = [self.error_message('required', field_name=key)]

        # Properties
        for key, child_schema in self.properties.items():
            if key not in value:
                if child_schema.has_default():
                    validated[key] = child_schema.default
                continue
            item = value[key]
            try:
                validated[key] = child_schema.validate(
                    item,
                    definitions=definitions,
                    allow_coerce=allow_coerce
                )
            except ValidationError as exc:
                errors[key] = exc.messages

        # Pattern properties
        if self.pattern_properties:
            for key in list(value.keys()):
                for pattern, child_schema in self.pattern_properties.items():
                    if isinstance(key, str) and re.search(pattern, key):
                        item = value[key]
                        try:
                            validated[key] = child_schema.validate(
                                item, definitions=definitions,
                                allow_coerce=allow_coerce
                            )
                        except ValidationError as exc:
                            errors[key] = exc.messages

        # Additional properties
        remaining = [
            key for key in value.keys()
            if key not in set(validated.keys()) | set(errors.keys())
        ]

        if self.additional_properties is True:
            for key in remaining:
                validated[key] = value[key]
        elif self.additional_properties is False:
            for key in remaining:
                errors[key] = [self.error_message('invalid_property')]
        elif self.additional_properties is not None:
            child_schema = self.additional_properties
            for key in remaining:
                item = value[key]
                try:
                    validated[key] = child_schema.validate(
                        item,
                        definitions=definitions,
                        allow_coerce=allow_coerce
                    )
                except ValidationError as exc:
                    errors[key] = exc.messages

        if errors:
            error_messages = []
            for key, messages in errors.items():
                for message in messages:
                    index = [key] if message.index is None else [key] + message.index
                    error_message = ErrorMessage(message.text, message.code, index)
                    error_messages.append(error_message)
            raise ValidationError(error_messages)

        return validated


class Array(Validator):
    errors = {
        'type': 'Must be an array.',
        'null': 'May not be null.',
        'empty': 'Must not be empty.',
        'exact_items': 'Must have {min_items} items.',
        'min_items': 'Must have at least {min_items} items.',
        'max_items': 'Must have no more than {max_items} items.',
        'additional_items': 'May not contain additional items.',
        'unique_items': 'This item is not unique.',
    }

    def __init__(self, items=None, additional_items=None, min_items=None,
                 max_items=None, unique_items=False, **kwargs):
        super().__init__(**kwargs)

        items = list(items) if isinstance(items, (list, tuple)) else items

        assert items is None or hasattr(items, 'validate') or (
            isinstance(items, list) and
            all(hasattr(i, 'validate') for i in items)
        )
        assert additional_items in (None, True, False) or hasattr(additional_items, 'validate')
        assert min_items is None or isinstance(min_items, int)
        assert max_items is None or isinstance(max_items, int)
        assert isinstance(unique_items, bool)

        self.items = items
        self.additional_items = additional_items
        self.min_items = min_items
        self.max_items = max_items
        self.unique_items = unique_items

    def validate(self, value, definitions=None, allow_coerce=False):
        if value is None and self.allow_null:
            return None
        elif value is None:
            self.error('null')
        elif not isinstance(value, list):
            self.error('type')

        definitions = self.get_definitions(definitions)
        validated = []

        if self.min_items is not None and self.min_items == self.max_items and len(value) != self.min_items:
            self.error('exact_items')
        if self.min_items is not None and len(value) < self.min_items:
            if self.min_items == 1:
                self.error('empty')
            self.error('min_items')
        elif self.max_items is not None and len(value) > self.max_items:
            self.error('max_items')
        elif isinstance(self.items, list) and (self.additional_items is False) and len(value) > len(self.items):
            self.error('additional_items')

        # Ensure all items are of the right type.
        errors = {}
        if self.unique_items:
            seen_items = Uniqueness()

        for pos, item in enumerate(value):
            try:
                if isinstance(self.items, list):
                    if pos < len(self.items):
                        item = self.items[pos].validate(
                            item,
                            definitions=definitions,
                            allow_coerce=allow_coerce
                        )
                    elif isinstance(self.additional_items, Validator):
                        item = self.additional_items.validate(
                            item,
                            definitions=definitions,
                            allow_coerce=allow_coerce
                        )
                elif self.items is not None:
                    item = self.items.validate(
                        item,
                        definitions=definitions,
                        allow_coerce=allow_coerce
                    )

                if self.unique_items:
                    if item in seen_items:
                        self.error('unique_items')
                    else:
                        seen_items.add(item)

                validated.append(item)
            except ValidationError as exc:
                errors[pos] = exc.messages

        if errors:
            error_messages = []
            for key, messages in errors.items():
                for message in messages:
                    index = [key] if message.index is None else [key] + message.index
                    error_message = ErrorMessage(message.text, message.code, index)
                    error_messages.append(error_message)
            raise ValidationError(error_messages)

        return validated


class Date(String):
    def __init__(self, **kwargs):
        super().__init__(format='date', **kwargs)


class Time(String):
    def __init__(self, **kwargs):
        super().__init__(format='time', **kwargs)


class DateTime(String):
    def __init__(self, **kwargs):
        super().__init__(format='datetime', **kwargs)


class Any(Validator):
    def validate(self, value, definitions=None, allow_coerce=False):
        # TODO: Validate value matches primitive types
        return value


class Union(Validator):
    errors = {
        'null': 'Must not be null.',
        'union': 'Must match one of the union types.'
    }

    def __init__(self, items, **kwargs):
        super().__init__(**kwargs)
        assert isinstance(items, list) and all(isinstance(i, Validator) for i in items)
        self.items = list(items)

    def validate(self, value, definitions=None, allow_coerce=False):
        if value is None and self.allow_null:
            return None
        elif value is None:
            self.error('null')

        for item in self.items:
            try:
                return item.validate(
                    value,
                    definitions=definitions,
                    allow_coerce=allow_coerce
                )
            except ValidationError:
                pass
        self.error('union')


class Ref(Validator):
    def __init__(self, ref, **kwargs):
        super().__init__(**kwargs)
        assert isinstance(ref, str)
        self.ref = ref

    def validate(self, value, definitions=None, allow_coerce=False):
        assert definitions is not None, 'Ref.validate() requires definitions'
        assert self.ref in definitions, 'Ref "%s" not in definitions' % self.ref

        child_schema = definitions[self.ref]
        return child_schema.validate(
            value,
            definitions=definitions,
            allow_coerce=allow_coerce
        )


class Uniqueness():
    """
    A set-like class that tests for uniqueness of primitive types.

    Ensures the `True` and `False` are treated as distinct from `1` and `0`,
    and coerces non-hashable instances that cannot be added to sets,
    into hashable representations that can.
    """
    TRUE = object()
    FALSE = object()

    def __init__(self):
        self._set = set()

    def __contains__(self, item):
        item = self.make_hashable(item)
        return item in self._set

    def add(self, item):
        item = self.make_hashable(item)
        self._set.add(item)

    def make_hashable(self, element):
        """
        Coerce a primitive into a uniquely hashable type, for uniqueness checks.
        """

        # Only primitive types can be handled.
        assert (element is None) or isinstance(element, (bool, int, float, str, list, dict))

        if element is True:
            # Need to make `True` distinct from `1`.
            return self.TRUE
        elif element is False:
            # Need to make `False` distinct from `0`.
            return self.FALSE
        elif isinstance(element, list):
            # Represent lists using a two-tuple of ('list', (item, item, ...))
            return ('list', tuple([
                self.make_hashable(item) for item in element
            ]))
        elif isinstance(element, dict):
            # Represent dicts using a two-tuple of ('dict', ((key, val), (key, val), ...))
            return ('dict', tuple([
                (self.make_hashable(key), self.make_hashable(value)) for key, value in element.items()
            ]))

        return element



import http.server
import json
import os
import shutil
import socketserver
import sys

import click

import apistar
from apistar.client import Client
from apistar.client.debug import DebugSession
from apistar.exceptions import (
    ClientError, ErrorResponse, ParseError, ValidationError
)


def _encoding_from_filename(filename):
    base, extension = os.path.splitext(filename)
    return {
        '.json': 'json',
        '.yml': 'yaml',
        '.yaml': 'yaml'
    }.get(extension)


def _echo_error(exc, content, verbose=False):
    if verbose:
        # Verbose output style.
        lines = content.splitlines()
        for message in reversed(exc.messages):
            error_str = ' ' * (message.position.column_no - 1)
            error_str += '^ '
            error_str += message.text
            error_str = click.style(error_str, fg='red')
            lines.insert(message.position.line_no, error_str)
        for line in lines:
            click.echo(line)

        click.echo()
    else:
        # Compact output style.
        for message in exc.messages:
            pos = message.position
            if message.code == 'required':
                index = message.index[:-1]
            else:
                index = message.index
            if index:
                fmt = '* %s (At %s, line %d, column %d.)'
                output = fmt % (message.text, index, pos.line_no, pos.column_no)
                click.echo(output)
            else:
                fmt = '* %s (At line %d, column %d.)'
                output = fmt % (message.text, pos.line_no, pos.column_no)
                click.echo(output)

    click.echo(click.style('✘ ', fg='red') + exc.summary)


def _copy_tree(src, dst, verbose=False):
    if not os.path.exists(dst):
        os.makedirs(dst)

    for name in os.listdir(src):
        srcname = os.path.join(src, name)
        dstname = os.path.join(dst, name)
        if os.path.isdir(srcname):
            _copy_tree(srcname, dstname, verbose=verbose)
        else:
            if verbose:
                click.echo(dstname)
            shutil.copy2(srcname, dstname)


def _load_config(options, verbose=False):
    if not os.path.exists('apistar.yml'):
        # If the config file is not used, then --path is required.
        if options['schema']['path'] is None:
            raise click.UsageError('Missing option "--path".')
        config = options

    else:
        with open('apistar.yml', 'rb') as config_file:
            content = config_file.read()

        try:
            config = apistar.validate(content, format='config', encoding="yaml")
        except (ParseError, ValidationError) as exc:
            click.echo('Errors in configuration file "apistar.yml":')
            _echo_error(exc, content, verbose=verbose)
            sys.exit(1)

        # Anything passed in 'options' should override the config file.
        # Ensure all values are populated with 'None' if unset.
        for section in options.keys():
            config.setdefault(section, {})
            for key, value in options[section].items():
                config[section].setdefault(key, None)
                if value is not None:
                    config[section][key] = value

    path = config['schema']['path']
    if not os.path.exists(path):
        raise click.UsageError('Schema file "%s" not found.' % path)

    if config['schema']['encoding'] is None:
        config['schema']['encoding'] = _encoding_from_filename(path)

    return config


FORMAT_SCHEMA_CHOICES = click.Choice(['openapi', 'swagger'])
FORMAT_ALL_CHOICES = click.Choice(['config', 'jsonschema', 'openapi', 'swagger'])
ENCODING_CHOICES = click.Choice(['json', 'yaml'])
THEME_CHOICES = click.Choice(['apistar', 'redoc', 'swaggerui'])


@click.group()
def cli():
    pass


@click.command()
@click.option('--path', type=click.Path(exists=True, dir_okay=False))
@click.option('--format', type=FORMAT_ALL_CHOICES)
@click.option('--encoding', type=ENCODING_CHOICES)
@click.option('--verbose', '-v', is_flag=True, default=False)
def validate(path, format, encoding, verbose):
    options = {
        'schema': {
            'path': path,
            'format': format,
            'encoding': encoding
        }
    }
    config = _load_config(options, verbose=verbose)

    path = config['schema']['path']
    format = config['schema']['format']
    encoding = config['schema']['encoding']

    with open(path, 'rb') as schema_file:
        content = schema_file.read()

    try:
        apistar.validate(content, format=format, encoding=encoding)
    except (ParseError, ValidationError) as exc:
        _echo_error(exc, content, verbose=verbose)
        sys.exit(1)

    success_summary = {
        'json': 'Valid JSON',
        'yaml': 'Valid YAML',
        'config': 'Valid APIStar config.',
        'jsonschema': 'Valid JSONSchema document.',
        'openapi': 'Valid OpenAPI schema.',
        'swagger': 'Valid Swagger schema.',
    }[format]
    click.echo(click.style('✓ ', fg='green') + success_summary)


@click.command()
@click.option('--path', type=click.Path(exists=True, dir_okay=False))
@click.option('--format', type=FORMAT_SCHEMA_CHOICES)
@click.option('--encoding', type=ENCODING_CHOICES)
@click.option('--output-dir', type=click.Path())
@click.option('--theme', type=THEME_CHOICES)
@click.option('--serve', is_flag=True, default=False)
@click.option('--verbose', '-v', is_flag=True, default=False)
def docs(path, format, encoding, output_dir, theme, serve, verbose):
    options = {
        'schema': {
            'path': path,
            'format': format,
            'encoding': encoding,
        },
        'docs': {
            'output_dir': output_dir,
            'theme': theme,
        }
    }
    config = _load_config(options, verbose=verbose)

    path = config['schema']['path']
    format = config['schema']['format']
    encoding = config['schema']['encoding']
    output_dir = config['docs']['output_dir']
    theme = config['docs']['theme']

    if output_dir is None:
        output_dir = 'build'
    if theme is None:
        theme = 'apistar'

    schema_filename = os.path.basename(path)
    schema_url = '/' + schema_filename
    with open(path, 'rb') as schema_file:
        content = schema_file.read()

    try:
        index_html = apistar.docs(content, format=format, encoding=encoding, schema_url=schema_url)
    except (ParseError, ValidationError) as exc:
        _echo_error(exc, content, verbose=verbose)
        sys.exit(1)

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    # Write 'index.html' to the docs.
    output_path = os.path.join(output_dir, 'index.html')
    if verbose:
        click.echo(output_path)
    output_file = open(output_path, 'w')
    output_file.write(index_html)
    output_file.close()

    # Write schema file to the docs.
    schema_path = os.path.join(output_dir, schema_filename)
    if verbose:
        click.echo(schema_path)
    shutil.copy2(path, schema_path)

    # Write static files to the docs.
    package_dir = os.path.dirname(apistar.__file__)
    static_dir = os.path.join(package_dir, 'themes', theme, 'static')
    _copy_tree(static_dir, output_dir, verbose=verbose)

    # All done.
    if serve:
        os.chdir(output_dir)
        addr = ("", 8000)
        handler = http.server.SimpleHTTPRequestHandler
        socketserver.TCPServer.allow_reuse_address = True
        with socketserver.TCPServer(addr, handler) as httpd:
            msg = 'Documentation available at "http://127.0.0.1:8000/" (Ctrl+C to quit)'
            click.echo(click.style('✓ ', fg='green') + msg)
            httpd.serve_forever()
    else:
        msg = 'Documentation built at "%s".'
        click.echo(click.style('✓ ', fg='green') + (msg % output_path))


@click.command()
@click.argument('operation')
@click.argument('params', nargs=-1)
@click.option('--path', type=click.Path(exists=True, dir_okay=False))
@click.option('--format', type=FORMAT_SCHEMA_CHOICES)
@click.option('--encoding', type=ENCODING_CHOICES)
@click.option('--verbose', '-v', is_flag=True, default=False)
@click.pass_context
def request(ctx, operation, params, path, format, encoding, verbose):
    options = {
        'schema': {
            'path': path,
            'format': format,
            'encoding': encoding
        }
    }
    config = _load_config(options, verbose=verbose)

    path = config['schema']['path']
    format = config['schema']['format']
    encoding = config['schema']['encoding']

    with open(path, 'rb') as schema_file:
        schema = schema_file.read()

    params = [param.partition('=') for param in params]
    params = dict([(key, value) for key, sep, value in params])

    session = ctx.obj

    if verbose:
        session = DebugSession(session)

    try:
        client = Client(schema, format=format, encoding=encoding, session=session)
    except (ParseError, ValidationError) as exc:
        _echo_error(exc, schema, verbose=verbose)
        sys.exit(1)

    try:
        result = client.request(operation, **params)
    except ClientError as exc:
        for message in exc.messages:
            if message.code == 'invalid_property':
                text = '* Invalid parameter "%s".' % exc.index[0]
            elif message.code == 'required':
                text = '* Missing required parameter "%s".' % exc.index[0]
            else:
                text = exc.text
            click.echo(text)
        click.echo(click.style('✘ ', fg='red') + 'Client error')
        sys.exit(1)
    except ErrorResponse as exc:
        click.echo(json.dumps(exc.content, indent=4))
        click.echo(click.style('✘ ', fg='red') + exc.title)
        sys.exit(1)
    click.echo(json.dumps(result, indent=4))


cli.add_command(docs)
cli.add_command(validate)
cli.add_command(request)



import yaml
from yaml.loader import SafeLoader

from apistar.exceptions import ErrorMessage, ParseError, Position
from apistar.tokenize.tokens import DictToken, ListToken, ScalarToken


def _get_position(content, index):
    return Position(
        line_no=content.count('\n', 0, index) + 1,
        column_no=index - content.rfind('\n', 0, index),
        index=index
    )


def tokenize_yaml(content):
    class CustomLoader(SafeLoader):
        pass

    def construct_mapping(loader, node):
        start = node.start_mark.index
        end = node.end_mark.index
        mapping = loader.construct_mapping(node)
        return DictToken(mapping, start, end - 1, content=content)

    def construct_sequence(loader, node):
        start = node.start_mark.index
        end = node.end_mark.index
        value = loader.construct_sequence(node)
        return ListToken(value, start, end - 1, content=content)

    def construct_scalar(loader, node):
        start = node.start_mark.index
        end = node.end_mark.index
        value = loader.construct_scalar(node)
        return ScalarToken(value, start, end - 1, content=content)

    def construct_int(loader, node):
        start = node.start_mark.index
        end = node.end_mark.index
        value = loader.construct_yaml_int(node)
        return ScalarToken(value, start, end - 1, content=content)

    def construct_float(loader, node):
        start = node.start_mark.index
        end = node.end_mark.index
        value = loader.construct_yaml_float(node)
        return ScalarToken(value, start, end - 1, content=content)

    def construct_bool(loader, node):
        start = node.start_mark.index
        end = node.end_mark.index
        value = loader.construct_yaml_bool(node)
        return ScalarToken(value, start, end - 1, content=content)

    def construct_null(loader, node):
        start = node.start_mark.index
        end = node.end_mark.index
        value = loader.construct_yaml_null(node)
        return ScalarToken(value, start, end - 1, content=content)

    CustomLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
        construct_mapping)

    CustomLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_SEQUENCE_TAG,
        construct_sequence)

    CustomLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_SCALAR_TAG,
        construct_scalar)

    CustomLoader.add_constructor(
        'tag:yaml.org,2002:int',
        construct_int)

    CustomLoader.add_constructor(
        'tag:yaml.org,2002:float',
        construct_float)

    CustomLoader.add_constructor(
        'tag:yaml.org,2002:bool',
        construct_bool)

    CustomLoader.add_constructor(
        'tag:yaml.org,2002:null',
        construct_null)

    assert isinstance(content, (str, bytes))

    if isinstance(content, bytes):
        content = content.decode('utf-8', 'ignore')

    if not content.strip():
        message = ErrorMessage(
            text='No content.',
            code='parse_error',
            position=Position(line_no=1, column_no=1, index=0)
        )
        raise ParseError(errors=[message], summary='Invalid YAML.')

    try:
        return yaml.load(content, CustomLoader)
    except (yaml.scanner.ScannerError, yaml.parser.ParserError) as exc:
        index = getattr(exc, 'index', 0)
        message = ErrorMessage(
            text=exc.problem + ".",
            code='parse_error',
            position=_get_position(content, index=index)
        )
        raise ParseError(messages=[message], summary='Invalid YAML.') from None



from apistar import validators

APISTAR_CONFIG = validators.Object(
    properties=[
        ('schema', validators.Object(
            properties=[
                ('path', validators.String()),
                ('format', validators.String(enum=['openapi', 'swagger'])),
                ('base_format', validators.String(enum=['json', 'yaml'])),
            ],
            additional_properties=False,
            required=['path', 'format']
        )),
        ('docs', validators.Object(
            properties=[
                ('output_dir', validators.String()),
                ('theme', validators.String(enum=['apistar', 'redoc', 'swaggerui'])),
            ],
            additional_properties=False,
        ))
    ],
    additional_properties=False,
    required=['schema'],
)



import collections
import sys

if sys.version_info < (3, 6):
    dict_type = collections.OrderedDict
else:
    dict_type = dict


try:
    import jinja2
except ImportError:
    jinja2 = None


try:
    import pygments
    from pygments.lexers import get_lexer_by_name
    from pygments.formatters import HtmlFormatter

    def pygments_highlight(text, lang, style):
        lexer = get_lexer_by_name(lang, stripall=False)
        formatter = HtmlFormatter(nowrap=True, style=style)
        return pygments.highlight(text, lexer, formatter)

    def pygments_css(style):
        formatter = HtmlFormatter(style=style)
        return formatter.get_style_defs('.highlight')

except ImportError:
    pygments = None

    def pygments_highlight(text, lang, style):
        return text

    def pygments_css(style):
        return None


try:
    # Ideally we subclass `_TemporaryFileWrapper` to present a clear __repr__
    # for downloaded files.
    from tempfile import _TemporaryFileWrapper

    class DownloadedFile(_TemporaryFileWrapper):
        basename = None

        def __repr__(self):
            state = "closed" if self.closed else "open"
            mode = "" if self.closed else " '%s'" % self.file.mode
            return "<DownloadedFile '%s', %s%s>" % (self.name, state, mode)

        def __str__(self):
            return self.__repr__()

except ImportError:
    # On some platforms (eg GAE) the private _TemporaryFileWrapper may not be
    # available, just use the standard `NamedTemporaryFile` function
    # in this case.
    import tempfile

    DownloadedFile = tempfile.NamedTemporaryFile



"""
This module provides a `DebugSession` requests session class, that echos
requests and responses to the console.

We use this for the `apistar request --verbose` console command.
"""
from urllib.parse import urlparse

import click
from requests.adapters import HTTPAdapter
from requests.sessions import Session


def expand_args(fmt, args):
    if args:
        return fmt % args
    return fmt


def debug_request(request):
    def request_echo(fmt, *args):
        click.echo(click.style('> ', fg='blue') + expand_args(fmt, args))

    request_echo(click.style('%s %s HTTP/1.1', bold=True), request.method, request.path_url)

    if 'host' not in request.headers:
        request_echo('Host: %s', urlparse(request.url).netloc)

    for key, value in sorted(request.headers.items()):
        request_echo('%s: %s', key.title(), value)

    if request.body:
        body_text = request.body
        if isinstance(body_text, bytes):
            body_text = body_text.decode('utf-8')
        request_echo('')
        for line in body_text.splitlines():
            request_echo(line)


def debug_response(response):
    def success_echo(fmt, *args):
        prompt = click.style('< ', fg='green')
        click.echo(prompt + expand_args(fmt, args))

    def failure_echo(fmt, *args):
        prompt = click.style('< ', fg='red')
        click.echo(prompt + expand_args(fmt, args))

    def info_echo(fmt, *args):
        prompt = click.style('< ', fg='yellow')
        click.echo(prompt + expand_args(fmt, args))

    response_class = ('%s' % response.status_code)[0] + 'xx'
    if response_class == '2xx':
        response_echo = success_echo
    elif response_class in ('4xx', '5xx'):
        response_echo = failure_echo
    else:
        response_echo = info_echo

    response_echo(click.style('%d %s', bold=True), response.status_code, response.reason)
    for key, value in sorted(response.headers.items()):
        response_echo('%s: %s', key.title(), value)
    if response.content:
        response_echo('')
        for line in response.text.splitlines():
            response_echo(line)

    click.echo()


class DebugAdapter(HTTPAdapter):
    def __init__(self, wrapped_session=None):
        self.session = Session() if wrapped_session is None else wrapped_session
        super().__init__()

    def send(self, request, **kwargs):
        debug_request(request)
        response = self.session.send(request, **kwargs)
        debug_response(response)
        return response


def DebugSession(wrapped_session=None):
    session = Session()
    session.mount('https://', DebugAdapter(wrapped_session))
    session.mount('http://', DebugAdapter(wrapped_session))
    return session



import os
import re

import jinja2

from apistar.exceptions import ErrorMessage, ValidationError
from apistar.schemas.config import APISTAR_CONFIG
from apistar.schemas.jsonschema import JSON_SCHEMA
from apistar.schemas.openapi import OPEN_API, OpenAPI
from apistar.schemas.swagger import SWAGGER, Swagger
from apistar.tokenize.tokenize_json import tokenize_json
from apistar.tokenize.tokenize_yaml import tokenize_yaml

__all__ = ['docs', 'parse', 'validate']


FORMAT_CHOICES = ['config', 'jsonschema', 'openapi', 'swagger', None]

# The regexs give us a best-guess for the encoding if none is specified.
# They check to see if the document looks like it is probably a YAML object or
# probably a JSON object. It'll typically be best to specify the encoding
# explicitly, but this should do for convenience.
INFER_YAML = re.compile(r'^([ \t]*#.*\n|---[ \t]*\n)*\s*[A-Za-z0-9_-]+[ \t]*:')
INFER_JSON = re.compile(r'^\s*{\s*"[A-Za-z0-9_-]+"\s*:')


def parse(content, encoding=None, validator=None):
    if encoding not in (None, "json", "yaml"):
        raise ValueError('encoding must be either "json" or "yaml"')

    if encoding is None:
        if INFER_YAML.match(content):
            encoding = "yaml"
        elif INFER_JSON.match(content):
            encoding = "json"
        else:
            message = ErrorMessage(
                text="Unable to guess if encoding is JSON or YAML. Use the 'encoding' argument.",
                code="unknown_encoding"
            )
            raise ValidationError(messages=[message])

    if encoding == "json":
        token = tokenize_json(content)
    else:
        token = tokenize_yaml(content)

    value = token.get_value()

    if validator is not None:
        try:
            value = validator.validate(value)
        except ValidationError as exc:
            for message in exc.messages:
                if message.code == 'required':
                    message.position = token.lookup_position(message.index[:-1])
                elif message.code in ['invalid_property', 'invalid_key']:
                    message.position = token.lookup_key_position(message.index)
                else:
                    message.position = token.lookup_position(message.index)
            exc.messages = sorted(exc.messages, key=lambda x: x.position.index)
            raise exc

    return (value, token)


def validate(schema, format=None, encoding=None):
    if format not in FORMAT_CHOICES:
        raise ValueError('format must be one of %s' % FORMAT_CHOICES)

    if isinstance(schema, (str, bytes)):
        value, token = parse(schema, encoding)
    elif isinstance(schema, dict):
        if encoding is not None:
            raise ValueError('encoding must be `None`.')
        value, token = schema, None
    else:
        raise ValueError('schema must either be a dict, or a string/bytestring.')

    if format is None:
        if isinstance(value, dict) and "openapi" in value and "swagger" not in value:
            format = "openapi"
        elif isinstance(value, dict) and "swagger" in value and "openapi" not in value:
            format = "swagger"
        else:
            message = ErrorMessage(
                text="Unable to determine schema format. Use the 'format' argument.",
                code='unknown_format'
            )
            raise ValidationError(messages=[message])

    validator = {
        'config': APISTAR_CONFIG,
        'jsonschema': JSON_SCHEMA,
        'openapi': OPEN_API,
        'swagger': SWAGGER
    }[format]

    if validator is not None:
        try:
            value = validator.validate(value)
        except ValidationError as exc:
            exc.summary = {
                'config': 'Invalid configuration file.',
                'jsonschema': 'Invalid JSONSchema document.',
                'openapi': 'Invalid OpenAPI schema.',
                'swagger': 'Invalid Swagger schema.',
            }[format]
            if token is not None:
                for message in exc.messages:
                    if message.code == 'required':
                        message.position = token.lookup_position(message.index[:-1])
                    elif message.code in ['invalid_property', 'invalid_key']:
                        message.position = token.lookup_key_position(message.index)
                    else:
                        message.position = token.lookup_position(message.index)
                exc.messages = sorted(exc.messages, key=lambda x: x.position.index)
            raise exc

    if format in ['openapi', 'swagger']:
        decoder = {
            'openapi': OpenAPI,
            'swagger': Swagger
        }[format]
        value = decoder().load(value)

    return value


def docs(schema, format=None, encoding=None, theme='apistar', schema_url=None, static_url=None):
    if format not in [None, 'openapi', 'swagger']:
        raise ValueError('format must be either "openapi" or "swagger"')

    document = validate(schema, format=format, encoding=encoding)

    loader = jinja2.PrefixLoader({
        theme: jinja2.PackageLoader('apistar', os.path.join('themes', theme, 'templates'))
    })
    env = jinja2.Environment(autoescape=True, loader=loader)

    if static_url is None:
        def static_url_func(path):
            return '/' + path.lstrip('/')
    elif isinstance(static_url, str):
        def static_url_func(path):
            return static_url.rstrip('/') + '/' + path.lstrip('/')
    else:
        static_url_func = static_url

    template = env.get_template(os.path.join(theme, 'index.html'))
    return template.render(
        document=document,
        langs=['javascript', 'python'],
        code_style=None,
        static_url=static_url_func,
        schema_url=schema_url
    )



class _ForceMultiPartDict(dict):
    """
    A dictionary that always evaluates as True.
    Allows us to force requests to use multipart encoding, even when no
    file parameters are passed.
    """
    def __bool__(self):
        return True

    def __nonzero__(self):
        return True


class BaseEncoder:
    media_type = None

    def encode(self, options, content):
        raise NotImplementedError()


class JSONEncoder:
    media_type = "application/json"

    def encode(self, options, content):
        options['json'] = content


class URLEncodedEncoder:
    media_type = "application/x-www-form-urlencoded"

    def encode(self, options, content):
        options['data'] = content


class MultiPartEncoder:
    media_type = "multipart/form-data"

    def encode(self, options, content):
        data = {}
        files = _ForceMultiPartDict()

        for key, value in content.items():
            if self.is_file(value):
                files[key] = value
            else:
                data[key] = value
        options['data'] = data
        options['files'] = files

    def is_file(self, item):
        if hasattr(item, '__iter__') and not isinstance(item, (str, list, tuple, dict)):
            # A stream object.
            return True
        return False



from collections import namedtuple

Position = namedtuple('Position', ['line_no', 'column_no', 'index'])


class ErrorMessage:
    def __init__(self, text, code, index=None, position=None):
        self.text = text
        self.code = code
        self.index = index
        self.position = position

    def __eq__(self, other):
        return (
            self.text == other.text and
            self.code == other.code and
            self.index == other.index and
            self.position == other.position
        )

    def __repr__(self):
        return "%s(%s, code=%s, index=%s, position=%s)" % (
            self.__class__.__name__,
            repr(self.text),
            repr(self.code),
            repr(self.index),
            repr(self.position)
        )


class DecodeError(Exception):
    def __init__(self, messages, summary=None):
        self.messages = messages
        self.summary = summary
        super().__init__(messages)


class ParseError(DecodeError):
    pass


class ValidationError(DecodeError):
    def as_dict(self):
        ret = {}
        for message in self.messages:
            lookup = ret
            if message.index:
                for key in message.index[:-1]:
                    lookup.setdefault(key, {})
                    lookup = lookup[key]
            key = message.index[-1] if message.index else None
            lookup[key] = message.text
        return ret


class ErrorResponse(Exception):
    """
    Raised when a client request results in an error response being returned.
    """
    def __init__(self, title, status_code, content):
        self.title = title
        self.status_code = status_code
        self.content = content


class ClientError(Exception):
    """
    Raised when a client is unable to fulfil an API request.
    """
    def __init__(self, messages):
        self.messages = messages
        super().__init__(messages)



"""
The tokenize package allows us to return nicely contextual error messages that
indicate where in a document an error has occurred.

It does this by parsing JSON or YAML and returning items that store both the
value and the position of each element in the parsed datastructure.
"""



from apistar.exceptions import Position


class Token():
    def __init__(self, value, start_index: int, end_index: int, content=None):
        self._value = value
        self.start_index = start_index
        self.end_index = end_index
        self._content = content

    def get_value(self):
        raise NotImplemented()  # pragma: nocover

    @property
    def start(self):
        return self._get_position(self.start_index)

    @property
    def end(self):
        return self._get_position(self.end_index)

    def lookup_position(self, keys=None):
        token = self
        for key in keys or []:
            token = token[key]
        return token.start

    def lookup_key_position(self, keys):
        token = self
        for key in keys[:-1]:
            token = token[key]
        return token.get_key(keys[-1]).start

    def _get_position(self, index):
        content = self._content[:index + 1]
        lines = content.splitlines()
        line_no = max(len(lines), 1)
        column_no = 1 if not lines else max(len(lines[-1]), 1)
        return Position(line_no, column_no, index)

    def __repr__(self):
        return '%s(%s, %d, %d)' % (
            self.__class__.__name__,
            repr(self._value),
            self.start_index,
            self.end_index
        )

    def __eq__(self, other):
        return (
            self.get_value() == other.get_value() and
            self.start_index == other.start_index and
            self.end_index == other.end_index
        )


class ScalarToken(Token):
    def __hash__(self):
        return hash(self._value)

    def get_value(self):
        return self._value


class DictToken(Token):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._keys = {k._value: k for k in self._value.keys()}
        self._values = {k._value: v for k, v in self._value.items()}

    def get_value(self):
        return {
            key_token.get_value(): value_token.get_value()
            for key_token, value_token in self._value.items()
        }

    def get_key(self, key):
        return self._keys[key]

    def __getitem__(self, key):
        return self._values[key]


class ListToken(Token):
    def get_value(self):
        return [
            token.get_value() for token in self._value
        ]

    def __getitem__(self, key):
        return self._value[key]



from urllib.parse import quote, urljoin, urlparse

import apistar
from apistar import exceptions, validators
from apistar.client import transports


class Client():
    def __init__(
        self, schema, format=None, encoding=None, auth=None, decoders=None,
        encoders=None, headers=None, session=None, allow_cookies=True
    ):
        self.document = apistar.validate(schema, format=format, encoding=encoding)
        self.transport = self.init_transport(auth, decoders, encoders, headers, session, allow_cookies)

    def init_transport(self, auth=None, decoders=None, encoders=None, headers=None, session=None, allow_cookies=True):
        return transports.HTTPTransport(
            auth=auth,
            decoders=decoders,
            encoders=encoders,
            headers=headers,
            session=session,
            allow_cookies=allow_cookies
        )

    def lookup_operation(self, operation_id: str):
        for item in self.document.walk_links():
            if item.link.name == operation_id:
                return item.link
        text = 'Operation ID "%s" not found in schema.' % operation_id
        message = exceptions.ErrorMessage(text=text, code='invalid-operation')
        raise exceptions.ClientError(messages=[message])

    def get_url(self, link, params):
        url = urljoin(self.document.url, link.url)

        scheme = urlparse(url).scheme.lower()

        if not scheme:
            text = "URL missing scheme '%s'." % url
            message = exceptions.ErrorMessage(text=text, code='invalid-url')
            raise exceptions.ClientError(messages=[message])

        if scheme not in self.transport.schemes:
            text = "Unsupported URL scheme '%s'." % scheme
            message = exceptions.ErrorMessage(text=text, code='invalid-url')
            raise exceptions.ClientError(messages=[message])

        for field in link.get_path_fields():
            value = str(params[field.name])
            if '{%s}' % field.name in url:
                url = url.replace('{%s}' % field.name, quote(value, safe=''))
            elif '{+%s}' % field.name in url:
                url = url.replace('{+%s}' % field.name, quote(value, safe='/'))

        return url

    def get_query_params(self, link, params):
        return {
            field.name: params[field.name]
            for field in link.get_query_fields()
            if field.name in params
        }

    def get_content_and_encoding(self, link, params):
        body_field = link.get_body_field()
        if body_field and body_field.name in params:
            return (params[body_field.name], link.encoding)
        return (None, None)

    def request(self, operation_id: str, **params):
        link = self.lookup_operation(operation_id)

        validator = validators.Object(
            properties={field.name: validators.Any() for field in link.fields},
            required=[field.name for field in link.fields if field.required],
            additional_properties=False
        )
        try:
            validator.validate(params)
        except exceptions.ValidationError as exc:
            raise exceptions.ClientError(messages=exc.messages) from None

        method = link.method
        url = self.get_url(link, params)
        query_params = self.get_query_params(link, params)
        (content, encoding) = self.get_content_and_encoding(link, params)

        return self.transport.send(
            method,
            url,
            query_params=query_params,
            content=content,
            encoding=encoding
        )



from requests.auth import AuthBase


class TokenAuthentication(AuthBase):
    def __init__(self, token, scheme='Bearer'):
        """
        * Use an unauthenticated client, and make a request to obtain a token.
        * Create an authenticated client using eg. `TokenAuthentication(token="<token>")`
        """
        self.token = token
        self.scheme = scheme

    def __call__(self, request):
        request.headers['Authorization'] = '%s %s' % (self.scheme, self.token)
        return request


class SessionAuthentication(AuthBase):
    """
    Enables session based login.

    * Make an initial request to obtain a CSRF token.
    * Make a login request.
    """
    safe_methods = ('GET', 'HEAD', 'OPTIONS', 'TRACE')

    def __init__(self, csrf_cookie_name=None, csrf_header_name=None):
        self.csrf_cookie_name = csrf_cookie_name
        self.csrf_header_name = csrf_header_name
        self.csrf_token = None

    def store_csrf_token(self, response, **kwargs):
        if self.csrf_cookie_name in response.cookies:
            self.csrf_token = response.cookies[self.csrf_cookie_name]

    def __call__(self, request):
        if self.csrf_token and self.csrf_header_name is not None and (request.method not in self.safe_methods):
            request.headers[self.csrf_header_name] = self.csrf_token
        if self.csrf_cookie_name is not None:
            request.register_hook('response', self.store_csrf_token)
        return request



import http

import requests

from apistar import exceptions
from apistar.client import decoders, encoders


class BlockAllCookies(http.cookiejar.CookiePolicy):
    """
    A cookie policy that rejects all cookies.
    Used to override the default `requests` behavior.
    """
    return_ok = set_ok = domain_return_ok = path_return_ok = lambda self, *args, **kwargs: False
    netscape = True
    rfc2965 = hide_cookie2 = False


class BaseTransport():
    schemes = None

    def send(self, method, url, query_params=None, content=None, encoding=None):
        raise NotImplementedError()


class HTTPTransport(BaseTransport):
    schemes = ['http', 'https']
    default_decoders = [
        decoders.JSONDecoder(),
        decoders.TextDecoder(),
        decoders.DownloadDecoder()
    ]
    default_encoders = [
        encoders.JSONEncoder(),
        encoders.MultiPartEncoder(),
        encoders.URLEncodedEncoder(),
    ]

    def __init__(self, auth=None, decoders=None, encoders=None, headers=None, session=None, allow_cookies=True):
        from apistar import __version__

        if session is None:
            session = requests.Session()
        if auth is not None:
            session.auth = auth
        if not allow_cookies:
            session.cookies.set_policy(BlockAllCookies())

        self.session = session
        self.decoders = list(decoders) if decoders else list(self.default_decoders)
        self.encoders = list(encoders) if encoders else list(self.default_encoders)
        self.headers = {
            'accept': ', '.join([decoder.media_type for decoder in self.decoders]),
            'user-agent': 'apistar %s' % __version__
        }
        if headers:
            self.headers.update({
                key.lower(): value
                for key, value in headers.items()
            })

    def send(self, method, url, query_params=None, content=None, encoding=None):
        options = self.get_request_options(query_params, content, encoding)
        response = self.session.request(method, url, **options)
        result = self.decode_response_content(response)

        if 400 <= response.status_code <= 599:
            title = '%d %s' % (response.status_code, response.reason)
            raise exceptions.ErrorResponse(
                title=title,
                status_code=response.status_code,
                content=result
            )

        return result

    def get_encoder(self, encoding):
        """
        Given the value of the encoding, return the appropriate encoder for
        handling the request content.
        """
        content_type = encoding.split(';')[0].strip().lower()
        main_type = content_type.split('/')[0] + '/*'
        wildcard_type = '*/*'

        for codec in self.encoders:
            if codec.media_type in (content_type, main_type, wildcard_type):
                return codec

        text = "Unsupported encoding '%s' for request." % encoding
        message = exceptions.ErrorMessage(text=text, code='cannot-encode-request')
        raise exceptions.ClientError(messages=[message])

    def get_decoder(self, content_type=None):
        """
        Given the value of a 'Content-Type' header, return the appropriate
        decoder for handling the response content.
        """
        if content_type is None:
            return self.decoders[0]

        content_type = content_type.split(';')[0].strip().lower()
        main_type = content_type.split('/')[0] + '/*'
        wildcard_type = '*/*'

        for codec in self.decoders:
            if codec.media_type in (content_type, main_type, wildcard_type):
                return codec

        text = "Unsupported encoding '%s' in response Content-Type header." % content_type
        message = exceptions.ErrorMessage(text=text, code='cannot-decode-response')
        raise exceptions.ClientError(messages=[message])

    def get_request_options(self, query_params=None, content=None, encoding=None):
        """
        Return the 'options' for sending the outgoing request.
        """
        options = {
            'headers': dict(self.headers),
            'params': query_params
        }

        if content is None:
            return options

        encoder = self.get_encoder(encoding)
        encoder.encode(options, content)
        return options

    def decode_response_content(self, response):
        """
        Given an HTTP response, return the decoded data.
        """
        if not response.content:
            return None

        content_type = response.headers.get('content-type')
        decoder = self.get_decoder(content_type)
        return decoder.decode(response)
